/**********************************************************************
Copyright (c) 2009 Stefan Seelmann and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors :
 ...
 ***********************************************************************/

package org.datanucleus.store.ldap.fieldmanager;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;

import javax.naming.directory.Attributes;
import javax.naming.ldap.LdapName;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.MetaDataManager;
import org.datanucleus.metadata.Relation;
import org.datanucleus.sco.SCOUtils;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.ldap.LDAPStoreManager;
import org.datanucleus.store.ldap.LDAPUtils;
import org.datanucleus.store.ldap.LDAPUtils.LocationInfo;
import org.datanucleus.util.ClassUtils;
import org.datanucleus.util.Localiser;

import org.datanucleus.store.StoreManager;

/**
 * Mapping strategy for hierarchical mapping of relationships.
 */
public class RelationByHierarchyStrategy extends AbstractMappingStrategy
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.datanucleus.store.ldap.Localisation", LDAPStoreManager.class
            .getClassLoader());

    protected int fieldNumber;

    protected ExecutionContext om;
    
    protected StoreManager storeMgr;

    protected ClassLoaderResolver clr;

    protected AbstractClassMetaData effectiveClassMetaData;

    protected boolean isFieldHierarchicalMapped;

    protected boolean isFieldParentOfHierarchicalMapping;

    protected RelationByHierarchyStrategy(StoreManager storeMgr, ObjectProvider sm, AbstractMemberMetaData mmd, Attributes attributes)
    {
        super(sm, mmd, attributes);
        this.fieldNumber = mmd.getAbsoluteFieldNumber();
        this.om = sm.getExecutionContext();
        this.storeMgr = storeMgr;
        this.clr = om.getClassLoaderResolver();
        this.effectiveClassMetaData = LDAPUtils.getEffectiveClassMetaData(mmd, sm.getExecutionContext().getMetaDataManager());
        this.isFieldHierarchicalMapped = isChildOfHierarchicalMapping(mmd, sm.getExecutionContext().getMetaDataManager());
        this.isFieldParentOfHierarchicalMapping = isParentOfHierarchicalMapping(mmd, sm.getExecutionContext().getMetaDataManager());
    }

    public Object fetch()
    {
        if (isFieldHierarchicalMapped)
        {
            switch (mmd.getRelationType(clr))
            {
                case Relation.ONE_TO_ONE_UNI :
                case Relation.ONE_TO_ONE_BI :

                    // hierarchical mapped one-one relationships are always dependent!
                    mmd.setDependent(true);
                    Object value = getHierarchicalMappedChild(mmd, sm);
                    return value;

                case Relation.ONE_TO_MANY_UNI :
                case Relation.ONE_TO_MANY_BI :

                    if (mmd.hasCollection())
                    {
                        Collection<Object> coll = getHierarchicalMappedChildren(mmd.getCollection().getElementType(), mmd.getMappedBy(),
                            mmd, sm);
                        return sm.wrapSCOField(fieldNumber, coll, false, false, true);
                    }
            }
        }

        if (isFieldParentOfHierarchicalMapping)
        {
            // this field is used for hierarchical N-1 relation, load the parent
            LdapName dn = LDAPUtils.getDistinguishedNameForObject(storeMgr, sm, true);
            LocationInfo locationInfo = LDAPUtils.getLocationInfo(sm.getClassMetaData());
            LdapName parentDn = LDAPUtils.getParentDistingueshedName(dn, locationInfo.suffix);

            Object value = null;
            if (locationInfo.dn == null || !locationInfo.dn.equals(parentDn))
            {
                value = LDAPUtils.getObjectByDN(storeMgr, om, type, parentDn.toString());
            }
            return value;
        }

        // TODO Localise this
        throw new NucleusException(
                "Cant obtain value for field " + mmd.getFullFieldName() + " since type=" + mmd.getTypeName() + " is not supported for this datastore");
    }

    private Object getHierarchicalMappedChild(AbstractMemberMetaData mmd, ObjectProvider sm)
    {
        Collection<Object> coll = getHierarchicalMappedChildren(mmd.getTypeName(), mmd.getMappedBy(), null, sm);
        if (coll.iterator().hasNext())
        {
            Object value = coll.iterator().next();
            return value;
        }
        else
        {
            return null;
        }
    }

    private Collection<Object> getHierarchicalMappedChildren(String type, String mappedBy, AbstractMemberMetaData mmd,
        ObjectProvider sm)
    {
        Collection<Object> coll;

        Class collectionClass;
        if (mmd == null)
        {
            collectionClass = ArrayList.class;
        }
        else
        {
            collectionClass = mmd.getType();
            collectionClass = SCOUtils.getContainerInstanceType(collectionClass, mmd.getOrderMetaData() != null);
        }

        try
        {
            coll = (Collection<Object>) collectionClass.newInstance();
            
            ExecutionContext om = sm.getExecutionContext();
            ClassLoaderResolver clr = om.getClassLoaderResolver();
            MetaDataManager metaDataMgr = sm.getExecutionContext().getMetaDataManager();
            AbstractClassMetaData childCmd = metaDataMgr.getMetaDataForClass(type, clr);
            LdapName base = LDAPUtils.getDistinguishedNameForObject(storeMgr, sm, true);

            List<Object> objects = LDAPUtils.getObjectsOfCandidateType(storeMgr, om, childCmd, base, null, true, false);
            coll.addAll(objects);
        }
        catch (InstantiationException e)
        {
            throw new NucleusException("Error in trying to create object of type " + collectionClass.getName(), e);
        }
        catch (IllegalAccessException e)
        {
            throw new NucleusException("Error in trying to create object of type " + collectionClass.getName(), e);
        }

        return coll;
    }

    public void insert(Object value)
    {
        if (value != null)
        {
            if (isFieldParentOfHierarchicalMapping)
            {
                // nothing to do
            }
            else if (isFieldHierarchicalMapped)
            {
                switch (mmd.getRelationType(clr))
                {
                    case Relation.ONE_TO_ONE_UNI :
                    case Relation.ONE_TO_ONE_BI :
                    {
                        LDAPUtils.markForPersisting(value, om);
                        break;
                    }
                    case Relation.ONE_TO_MANY_UNI :
                    case Relation.ONE_TO_MANY_BI :
                    {
                        if (mmd.hasCollection())
                        {
                            // 1-N (collection) relation
                            Collection c = (Collection) value;
                            for (Object pc : c)
                            {
                                LDAPUtils.markForPersisting(pc, om);
                            }
                        }
                        break;
                    }
                    default :
                    {
                        // TODO Localise this
                        throw new NucleusException("Field " + mmd.getFullFieldName() + " cannot be persisted because type=" + mmd
                                .getTypeName() + " with relation type " + mmd.getRelationType(clr) + " is not supported for this datastore");
                    }
                }
            }
            else
            {
                // TODO Localise this
                throw new NucleusException(
                        "Field " + mmd.getFullFieldName() + " cannot be persisted because type=" + mmd.getTypeName() + " is not supported for this datastore");
            }
        }
    }

    public void update(Object value)
    {
        if (isFieldParentOfHierarchicalMapping)
        {
            if (value == null)
            {
                LDAPUtils.markForDeletion(sm.getObject(), om);
            }
        }
        else if (isFieldHierarchicalMapped)
        {
            if (value != null)
            {
                String mappedBy = mmd.getMappedBy();
                switch (mmd.getRelationType(clr))
                {
                    case Relation.ONE_TO_ONE_UNI :
                    {
                        // Get the initial loaded object.
                        Object oldValue = getHierarchicalMappedChild(mmd, sm);
                        if (oldValue != null)
                        {
                            ObjectProvider oldValueSM = LDAPUtils.getStateManagerForObject(oldValue, om, false);
                            if (mustDelete(oldValueSM))
                            {
                                // delete only if the old value is still a child of this state manager
                                LDAPUtils.markForDeletion(oldValue, om);
                            }
                        }

                        // if new value.sm != null
                        ObjectProvider valueSM = LDAPUtils.getStateManagerForObject(value, om, false);
                        if (valueSM != null)
                        {
                            LdapName oldDn = LDAPUtils.getDistinguishedNameForObject(storeMgr, valueSM, true);
                            LdapName newDn = LDAPUtils.getDistinguishedNameForObject(storeMgr, valueSM);
                            LDAPUtils.markForRename(storeMgr, value, om, oldDn, newDn);
                            LDAPUtils.unmarkForDeletion(value, om);
                        }
                        // no break
                    }
                    case Relation.ONE_TO_ONE_BI :
                    {
                        LDAPUtils.markForPersisting(value, om);
                        LDAPUtils.unmarkForDeletion(value, om);
                        break;
                    }
                    case Relation.ONE_TO_MANY_UNI :
                    case Relation.ONE_TO_MANY_BI :
                    {
                        if (mmd.hasCollection())
                        {
                            Collection<Object> coll = (Collection<Object>) value;
                            // Get the collection with the initial loaded elements.
                            // It was saved by FetchFieldManager and is used to determine added and removed elements.
                            // This is just a quick and dirty implementation, could use backed SCO later.
                            
                            Collection<Object> oldColl = getHierarchicalMappedChildren(mmd.getCollection().getElementType(), mmd.getMappedBy(),
                                mmd, sm);
                            Collection<Object> toAdd = null;
                            Collection<Object> toRemove = null;
                            Class instanceType = mmd.getType();
                            if (List.class.isAssignableFrom(instanceType))
                            {
                                toAdd = new ArrayList<Object>(coll);
                                toRemove = new ArrayList<Object>(oldColl);
                            }
                            else
                            {
                                toAdd = new HashSet<Object>(coll);
                                toRemove = new HashSet<Object>(oldColl);
                            }

                            toAdd.removeAll(oldColl);
                            for (Object pc : toAdd)
                            {
                                LDAPUtils.markForPersisting(pc, om);
                                LDAPUtils.unmarkForDeletion(pc, om);
                            }

                            if (mmd.getRelationType(clr) == Relation.ONE_TO_MANY_UNI)
                            {
                                toRemove.removeAll(coll);
                                for (Object pc : toRemove)
                                {
                                    ObjectProvider pcSM = LDAPUtils.getStateManagerForObject(pc, om, false);
                                    if (mustDelete(pcSM))
                                    {
                                        // delete only if the old value is still a child of this state manager
                                        // System.out.println("deleteMap.put(4): " + pc);
                                        LDAPUtils.markForDeletion(pc, om);
                                    }
                                }
                            }
                            else if (mappedBy != null)
                            {
                                toRemove.removeAll(coll);
                                for (Object pc : toRemove)
                                {
                                    ObjectProvider pcSM = LDAPUtils.getStateManagerForObject(pc, om, false);
                                    if (mustDelete(pcSM, mappedBy))
                                    {
                                        // System.out.println("deleteMap.put(5): " + pc);
                                        LDAPUtils.markForDeletion(pc, om);
                                    }
                                }
                            }
                        }
                        break;
                    }
                    default :
                    {
                        // TODO Localise this
                        throw new NucleusException("Field " + mmd.getFullFieldName() + " cannot be persisted because type=" + mmd
                                .getTypeName() + " with relation type " + mmd.getRelationType(clr) + " is not supported for this datastore");
                    }
                }
            }
        }
        else
        {
            // TODO Localise this
            throw new NucleusException(
                    "Field " + mmd.getFullFieldName() + " cannot be persisted because type=" + mmd.getTypeName() + " is not supported for this datastore");
        }
    }

    /**
     * Checks if the given PC must be deleted.
     * @param pcSM the PC state manager
     * @return true if must delete
     */
    private boolean mustDelete(ObjectProvider pcSM)
    {
        LdapName dn = LDAPUtils.getDistinguishedNameForObject(storeMgr, sm, true);
        LdapName pcDn = LDAPUtils.getDistinguishedNameForObject(storeMgr, pcSM, true);
        boolean b = pcDn.startsWith(dn);
        return b;
    }

    /**
     * Checks if the given PC must be deleted.
     * @param pc The PC
     * @param mappedBy the mapped by field name
     * @return true if must delete
     */
    private boolean mustDelete(ObjectProvider pcSM, String mappedBy)
    {
        try
        {
            Object pc = pcSM.getObject();
            Class<? extends Object> clazz = pc.getClass();
            String methodName = ClassUtils.getJavaBeanGetterName(mappedBy, false);
            Method method = ClassUtils.getMethodForClass(clazz, methodName, null);
            Object fieldPC = method.invoke(pc);
            Object myPC = sm.getObject();
            if (fieldPC == null || fieldPC == myPC)
            {
                return true;
            }
        }
        catch (IllegalAccessException e)
        {
            e.printStackTrace();
        }
        catch (InvocationTargetException e)
        {
            e.printStackTrace();
        }
        return false;
    }

    public static boolean isChildOfHierarchicalMapping(AbstractMemberMetaData mmd, MetaDataManager mmgr)
    {
        String mappedBy = mmd.getMappedBy();
        if (mappedBy != null)
        {
            AbstractClassMetaData targetCmd = LDAPUtils.getEffectiveClassMetaData(mmd, mmgr);
            AbstractMemberMetaData mappedByMmd = targetCmd.getMetaDataForMember(mappedBy);
            if (mappedByMmd != null)
            {
                boolean isMappedByParent = isParentOfHierarchicalMapping(mappedByMmd, mmgr);
                return isMappedByParent;
            }
        }

        return false;
    }

    /**
     * Checks if the given field is the parent of hierarchical mapping. This is the case if the fields name is equal to
     * the parent field name defined in distinguished name.
     * @param mmd the field meta data
     * @param mmgr
     * @return
     */
    public static boolean isParentOfHierarchicalMapping(AbstractMemberMetaData mmd, MetaDataManager mmgr)
    {
        AbstractClassMetaData cmd = mmd.getAbstractClassMetaData();
        LocationInfo locationInfo = LDAPUtils.getLocationInfo(cmd);
        if (locationInfo.parentFieldName != null)
        {
            return mmd.getName().equals(locationInfo.parentFieldName);
        }

        return false;
    }

    @Override
    public List<String> getAttributeNames()
    {
        return Collections.EMPTY_LIST;
    }
}
